/*
  xcam.c - Display webcam image in X11 window

  Version 1.1b - 27.12.2001

  (c) Martin Grfe
*/

# include <stdio.h>
# include <stdlib.h>
# include <unistd.h>
# include <string.h>
# include <signal.h>
# include <fcntl.h>
# include <wait.h>
# include <sched.h>
# include <errno.h>
# include <sys/ipc.h>
# include <sys/shm.h>
# include <sys/types.h>
# include <sys/ioctl.h>
# include <xview/xview.h>
# include <xview/frame.h>
# include <xview/panel.h>
# include <xview/canvas.h>
# include <xview/xv_xrect.h>
# include <xview/svrimage.h>
# include <linux/videodev.h>


# ifdef DEBUG
#   define debug(a) printf a; fflush(stdout)
# else
#   define debug(a)
# endif


# define xcam_status (image->data[(long)(image->height)*image->bytes_per_line])
# define xcam_set_status(status) image->data[(long)(image->height)*image->bytes_per_line]=status
# define XCAM_INIT 0	/* status values */
# define XCAM_PAUSE 1
# define XCAM_RUN 2
# define XCAM_SNAPSHOT 3
# define xcam_rgb_image (&(image->data[(long)(image->height)*image->bytes_per_line+2]))
# define xcam_filter (image->data[(long)(image->height)*image->bytes_per_line+1])
# define xcam_set_filter(value) image->data[(long)(image->height)*image->bytes_per_line+1]=value
# define XCAM_NEW_FILTER 16


struct rgb24_pixel
 {
  unsigned char red, green, blue;
 };


/* global variables */

Display *display;
Window xwin;
GC gc;

XImage *image;	/* XImage for transfering image data using shared mem */

Panel_item brightness_slider, contrast_slider;

int video_fd;


/*------------------------------------------------------+
 |							|
 | copy_image()						|
 |							|
 |	copy the XImage to the canvas. This function is	|
 |	used as CANVAS_REPAINT_PROC and as signal	|
 |	handler for SIGUSR1.				|
 |							|
 +------------------------------------------------------*/
void copy_image(void)
 {
  debug(("xcam: copy_image(), xcam_status=%d\n", xcam_status));

  if (xcam_status == XCAM_INIT)
   {
    debug(("xcam: XDrawString()\n"));
    XDrawString(display, xwin, gc, 10, image->height/2,
      "Waiting for video data...", 25);
    xcam_set_status(XCAM_RUN);
    debug(("xcam: xcam_status=%d, copy_image() done.\n", xcam_status));
    return;
   }

  debug(("xcam: XPutImage()\n"));
  XPutImage(display, xwin, gc, image, 0, 0, 0, 0, image->width, image->height);

  if (xcam_status == XCAM_PAUSE)
    xcam_set_status(XCAM_RUN);
  else if (xcam_status == XCAM_SNAPSHOT)
    xcam_set_status(XCAM_PAUSE);

  debug(("xcam: xcam_status=%d, copy_image() done.\n", xcam_status));
  return;
 }


/*------------------------------------------------------+
 |							|
 | quit_proc(item, event)				|
 |							|
 |	function called when quit button is pressed.	|
 |							|
 +------------------------------------------------------*/
void quit_proc(Panel_item item, Event *event)
 {
  Frame frame;

  debug(("xcam: quit_proc()\n"));

  frame = xv_get(item, PANEL_CLIENT_DATA);
  signal(SIGUSR1, SIG_DFL);
  debug(("xcam: xv_destroy_safe(...)\n"));
  xv_destroy_safe(frame);
  debug(("xcam: quit_proc() done.\n"));
  return;
 }


/*------------------------------------------------------+
 |							|
 | close_proc(item, event)				|
 |							|
 |	function called when close button of subframe	|
 |	is pressed.					|
 |							|
 +------------------------------------------------------*/
void close_proc(Panel_item item, Event *event)
 {
  Frame subframe;

  debug(("xcam: close_proc()\n"));

  subframe = xv_get(item, PANEL_CLIENT_DATA);

  xv_set(subframe, FRAME_CMD_PIN_STATE, FRAME_CMD_PIN_OUT,
    XV_SHOW, FALSE, 0L);

  if (strcmp((char *)xv_get(subframe, FRAME_LABEL), "X-Cam: Enter file name") == 0)
    xcam_set_status(XCAM_RUN);

  debug(("xcam: close_proc() done.\n"));
  return;
 }


/*------------------------------------------------------+
 |							|
 | subframe_done_proc(cmd_frame)			|
 |							|
 |	function called when pushpin of subframe is	|
 |	pulled out.					|
 |							|
 +------------------------------------------------------*/
void subframe_done_proc(Frame subframe)
 {
  debug(("xcam: subframe_done_proc()\n"));

  xv_set(subframe, XV_SHOW, FALSE, 0L);

  xcam_set_status(XCAM_RUN);

  debug(("xcam_status=%d, subframe_done_proc() done.\n", xcam_status));
  return;
 }


/*------------------------------------------------------+
 |							|
 | save_image_proc(item, event)				|
 |							|
 |	function called when file name is entered or	|
 |	save button pressed.				|
 |							|
 +------------------------------------------------------*/
void save_image_proc(Panel_item item, Event *event)
 {
  Frame subframe;
  FILE *stream;

  debug(("xcam: save_image_proc()\n"));

  subframe = xv_get(item, PANEL_CLIENT_DATA);

  debug(("xcam: fopen(\"%s\", \"w\")\n", (char *)xv_get(item, PANEL_VALUE)));
  if ((stream = fopen((char *)xv_get(item, PANEL_VALUE), "w")) == NULL)
   {
    int error_no;
    static char buffer[80];

    error_no = errno;
    strcpy(buffer, "ERROR: ");
    strncat(buffer, strerror(error_no), 72);
    xv_set(subframe, FRAME_LEFT_FOOTER, buffer, 0L);
   }
  else
   {
    xv_set(subframe, FRAME_LEFT_FOOTER, "", 0L);

    debug(("xcam: fprintf(..., \"P6\\n%%d %%d\\n255\\n\", ...)\n"));
    fprintf(stream, "P6\n%d %d\n255\n", image->width, image->height);

    debug(("xcam: fwrite(...)\n"));
    if (image->depth == 24)
      fwrite(image->data, sizeof(struct rgb24_pixel),
        (long)(image->width)*(image->height), stream);
    else
      fwrite(xcam_rgb_image, sizeof(struct rgb24_pixel),
        (long)(image->width)*(image->height), stream);
    fclose(stream);

    xv_set(subframe, FRAME_CMD_PIN_STATE, FRAME_CMD_PIN_OUT,
      XV_SHOW, FALSE, 0L);

    xcam_set_status(XCAM_RUN);
   }

  debug(("xcam: xcam_status=%d, save_image_proc() done.\n", xcam_status));
  return;
 }


/*------------------------------------------------------+
 |							|
 | snap_proc(item, event)				|
 |							|
 |	function called when save button is pressed.	|
 |							|
 +------------------------------------------------------*/
void snap_proc(Panel_item item, Event *event)
 {
  Frame subframe, frame;

  debug(("xcam: snap_proc()\n"));

  if ((subframe = xv_get(item, PANEL_CLIENT_DATA)) == 0)
   {
    fprintf(stderr, "xcam: Can't get panel client data.\n");
    return;
   }

  if (xv_get(subframe, XV_SHOW) == TRUE)	/* subframe already open */
   {
    debug(("xcam: snap_proc() aborted.\n"));
    return;
   }

  if ((frame = xv_get(subframe, XV_OWNER)) == 0)
    fprintf(stderr, "xcam: Can't get parent frame handle.\n");

  xcam_set_status(XCAM_SNAPSHOT);

  while (xcam_status == XCAM_SNAPSHOT)	/* wait for parent proc to finish */
    sched_yield();
  debug(("xcam: xcam_status=%d, proceeding snap_proc()\n", xcam_status));

  if (frame != 0)
   {
    xv_set(subframe, XV_X, (int)xv_get(frame, XV_X)+40, 0L);
    xv_set(subframe, XV_Y, (int)xv_get(frame, XV_Y)+20, 0L);
   }
  xv_set(subframe, FRAME_LEFT_FOOTER, "", 0L);
  xv_set(subframe, FRAME_CMD_PIN_STATE, FRAME_CMD_PIN_IN,
    XV_SHOW, TRUE,
    0L);

  debug(("xcam: snap_proc() done.\n"));
  return;
 }


/*------------------------------------------------------+
 |							|
 | adjust_popup_proc(item, event)			|
 |							|
 |	function called when adjust button is pressed.	|
 |	Opens the Brightness/Contrast window.		|
 |							|
 +------------------------------------------------------*/
void adjust_popup_proc(Panel_item item, Event *event)
 {
  Frame frame, adjust_frame;
  struct video_picture video_pic;

  debug(("xcam: adjust_popup_proc()\n"));

  if ((adjust_frame = xv_get(item, PANEL_CLIENT_DATA)) == 0)
   {
    fprintf(stderr, "xcam: Can't get panel client data.\n");
    return;
   }

  if ((frame = xv_get(adjust_frame, XV_OWNER)) == 0)
    fprintf(stderr, "xcam: Can't get parent frame handle.\n");

  if (frame != 0)
   {
    xv_set(adjust_frame, XV_X, (int)xv_get(frame, XV_X)+250, 0L);
    xv_set(adjust_frame, XV_Y, (int)xv_get(frame, XV_Y)+10, 0L);
   }
  xv_set(adjust_frame, FRAME_CMD_PIN_STATE, FRAME_CMD_PIN_IN,
    XV_SHOW, TRUE,
    0L);

  if (ioctl(video_fd, VIDIOCGPICT, &video_pic) == -1)
   fprintf(stderr, "xcam: Can't get video picture.\n");
  else
   {
    xv_set(brightness_slider, PANEL_VALUE, (int)(video_pic.brightness/655.35+0.5), 0L);
    xv_set(contrast_slider, PANEL_VALUE, (int)(video_pic.contrast/655.35+0.5), 0L);
   }

  debug(("xcam: adjust_popup_proc() done.\n"));
  return;
 }


/*------------------------------------------------------+
 |							|
 | adjust_proc(item, event)				|
 |							|
 |	function called when brightness or contrast	|
 |	slider is changed.				|
 |							|
 +------------------------------------------------------*/
void adjust_proc(Panel_item item, Event *event)
 {
  struct video_picture video_pic;

  debug(("xcam: adjust_proc()\n"));

  if (ioctl(video_fd, VIDIOCGPICT, &video_pic) == -1)
    fprintf(stderr, "xcam: Can't get video picture.\n");
  else
   {
    video_pic.brightness = 0.5+655.35*(int)xv_get(brightness_slider, PANEL_VALUE);
    video_pic.contrast = 0.5+655.35*(int)xv_get(contrast_slider, PANEL_VALUE);
    if (ioctl(video_fd, VIDIOCSPICT, &video_pic) == -1)
      fprintf(stderr, "xcam: Can't set video picture.\n");
   }

  debug(("xcam: adjust_proc() done.\n"));
  return;
 }


/*------------------------------------------------------+
 |							|
 | filter_proc(item, event)				|
 |							|
 |	function called when filter level is changed.	|
 |							|
 +------------------------------------------------------*/
void filter_proc(Panel_item item, int value, Event *event)
 {
  debug(("xcam: filter_proc()\n"));

  if (value > 0)
    xcam_set_filter(value + XCAM_NEW_FILTER);	/* denote that filter has changed */
  else
    xcam_set_filter(0);

  debug(("xcam: xcam_filter=%d, filter_proc() done.\n", xcam_filter));
  return;
 }


/*------------------------------------------------------+
 |							|
 | deviation = colour_dev(rgbcolor1, rgbcolor2)		|
 |							|
 |	calculates the deviation between two RGB	|
 |	colours.					|
 |							|
 +------------------------------------------------------*/
double colour_dev(struct rgb24_pixel *color1, struct rgb24_pixel *color2)
 {
  double dev_red, dev_green, dev_blue, sum;

  dev_red = color1->red - color2->red;
  dev_green = color1->green - color2->green;
  dev_blue = color1->blue - color2->blue;
  sum = dev_red*dev_red + dev_green*dev_green + dev_blue*dev_blue;

  return(sum);
 }


/*==============================================================*/

int main(int argc, char *argv[])
 {
  int width, height, depth, screen_no, bytes_per_line,
    sh_mem_id, pid, old_width, old_height;
  Frame frame, subframe, adjust_frame;
  Canvas canvas;
  Panel main_panel, subframe_panel, adjust_panel;
  Panel_item quit_button, snap_button;
  Xv_Screen screen;
  Visual *visual;
  unsigned char *data, *pos;
  struct rgb24_pixel *rgb_ptr, *rgb_ptr2, *filter_buf;
  long length, l;
  struct video_window video_win;
  static char current_wd[44] = "Current path: ", buffer[32];
  Icon xcam_icon;
  Server_image icon_image, icon_mask_image, brightness_image, contrast_image;
# include "xcam.icon"
# include "xcam_mask.icon"
# include "xcam_brightness_image.icon"
# include "xcam_contrast_image.icon"


	/*- - - - - - - - - Passing arguments - - - - - - - - - -*/

  if ((argc == 2) && (strcmp(argv[1], "-h") == 0))
   {
    printf("X-Cam - Display WebCam image in an X11 window\n");
    printf("Version 1.1b - 2001 by Martin Grfe (AuM.Graefe@t-online.de)\n");
    printf("Usage: xcam [<xview options>] [sif | cif | vga]\n");
    printf("Currently supported visuals: 24 bpp, 16 bpp (5-6-5 bits for r-g-b)\n");
    return(0);
   }

  xv_init(XV_INIT_ARGC_PTR_ARGV, &argc, argv, 0L);

  if (argc > 2)
   {
    fprintf(stderr, "xcam: Bad arguments. Type 'xcam -h' for help.\n");
    return(1);
   }

  if (argc == 2)
    if (strcasecmp(argv[1], "sif") == 0)
     {
      width = 320;
      height = 240;
     }
    else if (strcasecmp(argv[1], "cif") == 0)
     {
      width = 352;
      height = 288;
     }
    else if (strcasecmp(argv[1], "vga") == 0)
     {
      width = 640;
      height = 480;
     }
    else
     {
      fprintf(stderr, "xcam: Bad argument '%s'. Type 'xcam -h' for help.\n",
        argv[1]);
      return(1);
     }

  getcwd(&(current_wd[14]), 29);


	/*- - - - - - - - Open Video4Linux device - - - - - - - - -*/

  debug(("xcam: open(\"/dev/video\", O_RDONLY)\n"));
  if ((video_fd = open("/dev/video", O_RDONLY)) == -1)
   {
    perror("xcam: Can't open device '/dev/video'");
    return(1);
   }


	/*- - - - - - - - Try to set width and height - - - - - - - - -*/

  debug(("xcam: ioctl(..., VIDIOCGWIN, ...)\n"));
  if (ioctl(video_fd, VIDIOCGWIN, &video_win) == -1)
   {
    if (argc == 2)	/* width and height are given */
      fprintf(stderr, "xcam: Can't get current video settings. Leaving settings unchanged.\n");
    else
     {
      fprintf(stderr, "xcam: Can't get current video settings. Suggesting size CIF (352x288).\n");
      width = 352;
      height = 288;
     }
   }
  else
   {
    if (argc == 2)
     {
      old_width = video_win.width;
      old_height = video_win.height;
      video_win.width = width;
      video_win.height = height;
      debug(("xcam: ioctl(..., VIDIOCSWIN, ...)\n"));
      if (ioctl(video_fd, VIDIOCSWIN, &video_win) == -1)
       {
	width = old_width;
	height = old_height;
	fprintf(stderr, "xcam: Can't set image size. Using current setting instead (%dx%d).\n", width, height);
       }
     }
    else
     {
      width = video_win.width;
      height = video_win.height;
     }
   }

  debug(("xcam: width=%d, height=%d\n", width, height));
  if ((width < 0) || (width > 800)|| (height < 0) || (height > 600))
   {
    fprintf(stderr, "xcam: Bad width or height (%d x %d)\n", width, height);
    return(1);
   }

	/*- - - - - - - - - - Create frames - - - - - - - - - - -*/

  sprintf(buffer, "Image size: %d x %d", width, height);

  frame = (Frame)xv_create(0, FRAME,			/* basic frame */
    FRAME_LABEL, "X-Cam V1.1b",
    FRAME_SHOW_RESIZE_CORNER, FALSE,
    FRAME_LEFT_FOOTER, buffer, 
    FRAME_SHOW_FOOTER, TRUE,
    0L);
  if (frame == 0)
   {
    fprintf(stderr, "xcam: Can't create frame.\n");
    return(1);
   }

  subframe = (Frame)xv_create(frame, FRAME_CMD,		/* snapshot frame */
    FRAME_LABEL, "X-Cam: Enter file name",
    FRAME_SHOW_FOOTER, TRUE,
    FRAME_DONE_PROC, subframe_done_proc,
    0L);
  if (subframe == 0)
   {
    fprintf(stderr, "xcam: Can't create frame.\n");
    return(1);
   }

  subframe_panel = (Panel)xv_get(subframe, FRAME_CMD_PANEL);
  xv_set(subframe_panel, PANEL_LAYOUT, PANEL_VERTICAL, 0L);

  xv_create(subframe_panel, PANEL_MESSAGE,
    PANEL_LABEL_STRING, current_wd,
    0L);

  xv_create(subframe_panel, PANEL_TEXT,
    PANEL_LABEL_STRING, "File name: ",
    PANEL_VALUE_DISPLAY_WIDTH, 180,
    PANEL_NOTIFY_PROC, save_image_proc,
    PANEL_CLIENT_DATA, subframe,
    0L);

  xv_create(subframe_panel, PANEL_BUTTON,
    PANEL_LABEL_STRING, "Close",
    PANEL_NOTIFY_PROC, close_proc,
    PANEL_CLIENT_DATA, subframe,
    0L);

  window_fit(subframe_panel);
  window_fit(subframe);

  adjust_frame = (Frame)xv_create(frame, FRAME_CMD,	/* adjust frame */
    FRAME_LABEL, "X-Cam: Adjust",
    0L);
  if (adjust_frame == 0)
   {
    fprintf(stderr, "xcam: Can't create frame.\n");
    return(1);
   }

  adjust_panel = (Panel)xv_get(adjust_frame, FRAME_CMD_PANEL);
  xv_set(adjust_panel, PANEL_LAYOUT, PANEL_VERTICAL, 0L);

  brightness_image = (Server_image)xv_create(0L, SERVER_IMAGE,
    XV_WIDTH, xcam_brightness_image_width,
    XV_HEIGHT, xcam_brightness_image_height,
    SERVER_IMAGE_BITS, xcam_brightness_image_bits,
    0L);
  if (brightness_image == 0)
   {
    fprintf(stderr, "xcam: Can't create image.\n");
    brightness_slider = xv_create(adjust_panel, PANEL_SLIDER,
      PANEL_LABEL_STRING, "Brightness",
      PANEL_ITEM_Y, 10,
      PANEL_MIN_VALUE, 0,
      PANEL_MAX_VALUE, 100,
      PANEL_TICKS, 5,
      PANEL_MIN_TICK_STRING, "0%",
      PANEL_MAX_TICK_STRING, "100%",
      PANEL_SHOW_RANGE, FALSE,
      PANEL_DIRECTION, PANEL_HORIZONTAL,
      PANEL_SHOW_VALUE, FALSE,
      PANEL_NOTIFY_PROC, adjust_proc,
      0L);
   }
  else
   {
    brightness_slider = xv_create(adjust_panel, PANEL_SLIDER,
      PANEL_LABEL_IMAGE, brightness_image,
      PANEL_ITEM_Y, 10,
      PANEL_MIN_VALUE, 0,
      PANEL_MAX_VALUE, 100,
      PANEL_TICKS, 5,
      PANEL_MIN_TICK_STRING, "0%",
      PANEL_MAX_TICK_STRING, "100%",
      PANEL_SHOW_RANGE, FALSE,
      PANEL_DIRECTION, PANEL_HORIZONTAL,
      PANEL_SHOW_VALUE, FALSE,
      PANEL_NOTIFY_PROC, adjust_proc,
      0L);
   }

  contrast_image = (Server_image)xv_create(0L, SERVER_IMAGE,
    XV_WIDTH, xcam_contrast_image_width,
    XV_HEIGHT, xcam_contrast_image_height,
    SERVER_IMAGE_BITS, xcam_contrast_image_bits,
    0L);
  if (contrast_image == 0)
   {
    fprintf(stderr, "xcam: Can't create image.\n");
    contrast_slider = xv_create(adjust_panel, PANEL_SLIDER,
      PANEL_LABEL_STRING, "Contrast",
      PANEL_MIN_VALUE, 0,
      PANEL_MAX_VALUE, 100,
      PANEL_TICKS, 5,
      PANEL_MIN_TICK_STRING, "0%",
      PANEL_MAX_TICK_STRING, "100%",
      PANEL_SHOW_RANGE, FALSE,
      PANEL_DIRECTION, PANEL_HORIZONTAL,
      PANEL_SHOW_VALUE, FALSE,
      PANEL_NOTIFY_PROC, adjust_proc,
      0L);
   }
  else
   {
    contrast_slider = xv_create(adjust_panel, PANEL_SLIDER,
      PANEL_LABEL_IMAGE, contrast_image,
      PANEL_MIN_VALUE, 0,
      PANEL_MAX_VALUE, 100,
      PANEL_TICKS, 5,
      PANEL_MIN_TICK_STRING, "0%",
      PANEL_MAX_TICK_STRING, "100%",
      PANEL_SHOW_RANGE, FALSE,
      PANEL_DIRECTION, PANEL_HORIZONTAL,
      PANEL_SHOW_VALUE, FALSE,
      PANEL_NOTIFY_PROC, adjust_proc,
      0L);
   }

  xv_create(adjust_panel, PANEL_BUTTON,
    PANEL_LABEL_STRING, "Close",
    PANEL_NOTIFY_PROC, close_proc,
    PANEL_CLIENT_DATA, adjust_frame,
    0L);

  window_fit(adjust_panel);
  window_fit(adjust_frame);

  display = (Display *)xv_get(frame, XV_DISPLAY);
  screen = (Xv_Screen)xv_get(frame, XV_SCREEN);
  visual = (Visual *)xv_get(frame, XV_VISUAL);
  screen_no = (int)xv_get(screen, SCREEN_NUMBER);
  depth = DefaultDepth(display, screen_no);
  gc = DefaultGC(display, DefaultScreen(display));

  if ((depth != 16) && (depth != 24))
   {
    fprintf(stderr, "xcam: Unsupported depth (%d bpp)."
            " Please use 16 or 24 bpp.\n", depth);
    return(1);
   }

  if ((depth == 16) & ((visual->blue_mask != 0x001f)
       || (visual->green_mask != 0x07e0) || (visual->red_mask != 0xf800)))
   {
    fprintf(stderr, "xcam: Unsupported visual.\n");
    return(1);
   }

  bytes_per_line = (7L+width*depth)>>3;


	/*- - - - - - - - Create panel with buttons - - - - - - - - -*/

  main_panel = (Panel)xv_create(frame, PANEL,
    XV_WIDTH, width+2,
    PANEL_LAYOUT, PANEL_HORIZONTAL,
    0L);
  if (main_panel == 0)
   {
    fprintf(stderr, "xcam: Can't create panel.\n");
    return(1);
   }

  quit_button = (Panel_item)xv_create(main_panel, PANEL_BUTTON,
    PANEL_LABEL_STRING, "Quit",
    PANEL_NOTIFY_PROC, quit_proc,
    PANEL_CLIENT_DATA, frame,
    0L);
  if (quit_button == 0)
    fprintf(stderr, "xcam warning: Can't create button.\n");

  xv_create(main_panel, PANEL_BUTTON,
    PANEL_LABEL_STRING, "Adjust",
    PANEL_NOTIFY_PROC, adjust_popup_proc,
    PANEL_CLIENT_DATA, adjust_frame,
    0L);

  snap_button = (Panel_item)xv_create(main_panel, PANEL_BUTTON,
    PANEL_LABEL_STRING, "Snapshot",
    PANEL_NOTIFY_PROC, snap_proc,
    PANEL_CLIENT_DATA, subframe,
    0L);
  if (snap_button == 0)
    fprintf(stderr, "xcam warning: Can't create button.\n");

  xv_create(main_panel, PANEL_CHOICE_STACK,
    PANEL_LAYOUT, PANEL_HORIZONTAL,
    PANEL_LABEL_STRING, "Filter level:",
    PANEL_CHOICE_STRINGS, "off", " 1 ", " 2 ", NULL,
    PANEL_VALUE, 0,
    PANEL_NOTIFY_PROC, filter_proc,
    0L);

  window_fit_height(main_panel);
  

	/*- - - - - - - - - Create canvas - - - - - - - - - -*/

  canvas = (Canvas)xv_create(frame, CANVAS,
    CANVAS_AUTO_EXPAND, FALSE,
    CANVAS_AUTO_SHRINK, FALSE,
    CANVAS_WIDTH, width,
    CANVAS_HEIGHT, height,
    XV_WIDTH, width+2,
    XV_HEIGHT, height+2,
    XV_X, 0,
    XV_Y, xv_get(main_panel, XV_HEIGHT),
    CANVAS_REPAINT_PROC, copy_image,
    0L);
  if (canvas == 0)
   {
    fprintf(stderr, "xcam: Can't create canvas.\n");
    return(1);
   }

  xwin = xv_get(canvas_paint_window(canvas), XV_XID);


	/*- - - - - - - - - - Create Icon - - - - - - - - - -*/

  icon_image = (Server_image)xv_create(0, SERVER_IMAGE,
    XV_WIDTH, xcam_icon_width,
    XV_HEIGHT, xcam_icon_height,
    SERVER_IMAGE_BITS, xcam_icon_bits,
    0L);
  if (icon_image == 0)
    fprintf(stderr, "xcam: Can't create icon image.\n");
  else
   {
    icon_mask_image = (Server_image)xv_create(0, SERVER_IMAGE,
      XV_WIDTH, xcam_icon_mask_width,
      XV_HEIGHT, xcam_icon_mask_height,
      SERVER_IMAGE_BITS, xcam_icon_mask_bits,
      0L);
    if (icon_mask_image == 0)
      fprintf(stderr, "xcam: Can't create icon mask.\n");
    else
     {
      xcam_icon = (Icon)xv_create(frame, ICON,
	ICON_IMAGE, icon_image,
	ICON_MASK_IMAGE, icon_mask_image,
        ICON_WIDTH, xcam_icon_width,
        ICON_HEIGHT, xcam_icon_height,
	ICON_LABEL, "X-Cam",
	0L);
      if (xcam_icon == 0)
        fprintf(stderr, "xcam: Can't create frame icon.\n");
      else
        xv_set(frame, FRAME_ICON, xcam_icon, 0L);
     }
   }


	/*- - - - - - - - - allocating filter memory - - - - - - - - - -*/

  debug(("xcam: malloc(%ld)\n", (long)width*height*3));
  if ((filter_buf = (struct rgb24_pixel *)malloc((size_t)width*height*3)) == NULL)
   {
    fprintf(stderr, "xcam: Can't allocate memory.\n");
    return(1);
   }


	/*- - - - - - - - - attaching shared memory - - - - - - - - - -*/

  if (depth == 24)
    length = (size_t)bytes_per_line*height+2;
  else		/* for 16 bpp keep full color and reduced color image */
    length = (size_t)bytes_per_line*height + 2 + width*height*3;

  debug(("xcam: shmget(IPC_PRIVATE, %ld, SHM_R | SHM_W)\n", length));
  if ((sh_mem_id = shmget(IPC_PRIVATE, length, SHM_R | SHM_W)) < 0)
   {
    fprintf(stderr, "xcam: Can't get shared memory (shmget failed).\n");
    return(1);
   }
  if ((data = (char *)shmat(sh_mem_id, 0, 0)) == (void *)-1)
   {
    fprintf(stderr, "xcam: Can't attach shared memory (shmat failed).\n");
    shmctl(sh_mem_id, IPC_RMID, 0);
    return(1);
   }


	/*- - - - - - - - - Create XImage - - - - - - - - - -*/

  debug(("xcam: XCreateImage(depth=%d, width=%d, height=%d, bytes_per_line=%d)\n", depth, width, height, bytes_per_line));
  image = XCreateImage(display, visual, depth, ZPixmap, 0, data,
    width, height, 16, bytes_per_line);

  image->byte_order = MSBFirst;


	/*- - - - - - - Create Child Process (GUI) - - - - - - - -*/

  xcam_set_status(XCAM_INIT);
  xcam_set_filter(0);

  debug(("xcam: fork()\n"));
  if ((pid = fork()) == 0)
   {
    int p_pid;

    signal(SIGUSR1, (void *)copy_image);	/* install new signal handler */

    window_fit(frame);
    xv_main_loop(frame);
    xcam_set_status(XCAM_RUN);		/* to prevent parent proccess from hanging */

    p_pid = getppid();
    kill(SIGCHLD, p_pid);
    sleep(1);
    kill(SIGKILL, p_pid);		/* if read() in parent process hangs */

    return(0);
   }


	/*- - - - - - loop for reading images from web-cam - - - - - - -*/

  while (waitpid(pid, NULL, WNOHANG) != pid)
   {
    unsigned char filter_level;

    while ((xcam_status == XCAM_PAUSE) || (xcam_status == XCAM_INIT))
      sched_yield();

    length = 3L*width*height;

    if ((filter_level = xcam_filter) != 0)
      pos = (char *)filter_buf;
    else
      if (depth == 24)
	pos = data;
      else
	pos = xcam_rgb_image;

    debug(("xcam: read()\n"));
    while (length > 0)			/* read image from web-cam */
      if ((l = read(video_fd, pos, length)) < 0)
       {
	perror("xcam: Can't read from device '/dev/video'");
	kill(pid, SIGKILL);
	close(video_fd);
	shmctl(sh_mem_id, IPC_RMID, 0);
	shmdt(data);
	return(1);
       }
      else
       {
	length -= l;
	pos += l;
       }

    if (filter_level != 0)		/* do filtering if activated */
     {
      double scale1, scale2;

      debug(("xcam: running filter\n"));
      rgb_ptr = filter_buf;
      if (depth == 24)
	rgb_ptr2 = (struct rgb24_pixel *)data;
      else
	rgb_ptr2 = (struct rgb24_pixel *)xcam_rgb_image;

      if (filter_level & XCAM_NEW_FILTER)	/* new filter level */
       {
        xcam_set_filter(filter_level - XCAM_NEW_FILTER);
	for (l=0; l<(long)width*height; l++)
	 {
	  rgb_ptr2->red = rgb_ptr->red;
	  rgb_ptr2->green = rgb_ptr->green;
	  rgb_ptr2->blue = rgb_ptr->blue;
	  rgb_ptr++;
	  rgb_ptr2++;
	 }
       }
      else
       {
        scale1 = 1.0/(filter_level+1);
        scale2 = 1.0-scale1;
	for (l=0; l<(long)width*height; l++)
	 {
	  if (colour_dev(rgb_ptr, rgb_ptr2) < 1000)
	   {
	    rgb_ptr2->red = rgb_ptr->red*scale1 + rgb_ptr2->red*scale2 + 0.5;
	    rgb_ptr2->green = rgb_ptr->green*scale1 + rgb_ptr2->green*scale2 + 0.5;
	    rgb_ptr2->blue = rgb_ptr->blue*scale1 + rgb_ptr2->blue*scale2 + 0.5;
	   }
	  else
	   {
	    rgb_ptr2->red = rgb_ptr->red;
	    rgb_ptr2->green = rgb_ptr->green;
	    rgb_ptr2->blue = rgb_ptr->blue;
	   }
	  rgb_ptr++;
	  rgb_ptr2++;
	 }
       }
     }

    if (depth == 16)			/* for 64k colors map bits */
     {
      debug(("xcam: converting image to 16 bpp\n"));
      pos = data;
      rgb_ptr = (struct rgb24_pixel *)xcam_rgb_image;
      for (l=0; l<(long)width*height; l++)
       {
	*(pos++) = (rgb_ptr->red & 0xf8) | (rgb_ptr->green >> 5);
	*(pos++) = ((rgb_ptr->green << 3) & 0xe0) | (rgb_ptr->blue >> 3);
	rgb_ptr++;
       }
     }

    if (xcam_status == XCAM_RUN)
     {
      xcam_set_status(XCAM_PAUSE);
      debug(("xcam: kill(..., SIGUSR1)\n"));
      kill(pid, SIGUSR1);		/* start copy_image() of GUI process */
     }
    else if (xcam_status == XCAM_SNAPSHOT)
     {
      debug(("xcam: kill(..., SIGUSR1)\n"));
      kill(pid, SIGUSR1);		/* start copy_image() of GUI process */
      while (xcam_status == XCAM_SNAPSHOT)
        sched_yield();
     }
   }

  close(video_fd);
  free(filter_buf);
  shmctl(sh_mem_id, IPC_RMID, 0);
  shmdt(data);

  return(0);
 }
